#
# Copyright 2024 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import os
import webtest

from aodh.api import app
from aodh.storage import models
from aodh.tests import constants
from aodh.tests.functional.api import v2


def getTestAlarm(alarm_id, project_id, user_id):
    return models.Alarm(name='name1',
                        type='gnocchi_aggregation_by_metrics_threshold',
                        enabled=True,
                        alarm_id=alarm_id,
                        description='a',
                        state='insufficient data',
                        state_reason='Not evaluated',
                        severity='critical',
                        state_timestamp=constants.MIN_DATETIME,
                        timestamp=constants.MIN_DATETIME,
                        ok_actions=[],
                        insufficient_data_actions=[],
                        alarm_actions=[],
                        repeat_actions=True,
                        user_id=user_id,
                        project_id=project_id,
                        time_constraints=[dict(name='testcons',
                                               start='0 11 * * *',
                                               duration=300)],
                        rule=dict(comparison_operator='gt',
                                  threshold=2.0,
                                  aggregation_method='mean',
                                  evaluation_periods=60,
                                  granularity=1,
                                  metrics=[
                                      '41869681-5776-46d6-91ed-cccc43b6e4e3',
                                      'a1fb80f4-c242-4f57-87c6-68f47521059e'
                                  ])
                        )


class TestMetrics(v2.FunctionalTest):
    def setUp(self):
        super().setUp()
        self.project_id = "some_project_id"
        self.project_id2 = "some_project_id2"
        self.alarm_id = "some_alarm_id"
        self.alarm_id2 = "some_alarm_id2"
        self.user_id = "some_user_id"
        self.role = "reader"
        self.auth_headers = {'X-User-Id': self.user_id,
                             'X-Project-Id': self.project_id,
                             'X-Roles': self.role}
        self.alarm_conn.create_alarm(getTestAlarm(
            self.alarm_id,
            self.project_id,
            self.user_id)
        )
        self.alarm_conn.create_alarm(getTestAlarm(
            self.alarm_id2,
            self.project_id2,
            self.user_id)
        )
        self.alarm_conn.increment_alarm_counter(
            self.alarm_id,
            self.project_id,
            "ok"
        )
        self.alarm_conn.increment_alarm_counter(
            self.alarm_id,
            self.project_id,
            "insufficient_data"
        )
        self.alarm_conn.increment_alarm_counter(
            self.alarm_id,
            self.project_id,
            "insufficient_data"
        )
        self.alarm_conn.increment_alarm_counter(
            self.alarm_id2,
            self.project_id2,
            "alarm"
        )

    def test_get_all_metrics_inside_project(self):
        expected = {
            "evaluation_results":
            [{
                "alarm_id": self.alarm_id,
                "project_id": self.project_id,
                "state_counters": {
                    "ok": 1,
                    "insufficient data": 2,
                    "alarm": 0
                }
            }]
        }
        metrics = self.get_json(
            '/metrics',
            headers=self.auth_headers,
        )
        self.assertEqual(expected, metrics)

    def test_get_all_metrics_forbidden(self):
        pf = os.path.abspath('aodh/tests/functional/api/v2/policy.yaml-test')
        self.CONF.set_override('policy_file', pf, group='oslo_policy')
        self.CONF.set_override('auth_mode', None, group='api')
        self.app = webtest.TestApp(app.load_app(self.CONF))

        response = self.get_json('/metrics',
                                 expect_errors=True,
                                 status=403,
                                 headers=self.auth_headers)
        faultstring = 'RBAC Authorization Failed'
        self.assertEqual(403, response.status_code)
        self.assertEqual(faultstring,
                         response.json['error_message']['faultstring'])

    def test_get_all_metrics_all_projects(self):
        auth_headers = copy.copy(self.auth_headers)
        auth_headers['X-Roles'] = 'admin'
        expected = {
            "evaluation_results": [{
                "alarm_id": self.alarm_id,
                "project_id": self.project_id,
                "state_counters": {
                    "ok": 1,
                    "insufficient data": 2,
                    "alarm": 0
                }
            }, {
                "alarm_id": self.alarm_id2,
                "project_id": self.project_id2,
                "state_counters": {
                    "ok": 0,
                    "insufficient data": 0,
                    "alarm": 1
                }
            }]
        }
        metrics = self.get_json(
            '/metrics?all_projects=true',
            headers=auth_headers,
        )
        self.assertEqual(expected, metrics)

    def test_get_all_metrics_all_projects_forbidden(self):
        pf = os.path.abspath('aodh/tests/functional/api/v2/policy.yaml-test')
        self.CONF.set_override('policy_file', pf, group='oslo_policy')
        self.CONF.set_override('auth_mode', None, group='api')
        self.app = webtest.TestApp(app.load_app(self.CONF))

        response = self.get_json('/metrics?all_projects=true',
                                 expect_errors=True,
                                 status=403,
                                 headers=self.auth_headers)
        faultstring = 'RBAC Authorization Failed'
        self.assertEqual(403, response.status_code)
        self.assertEqual(faultstring,
                         response.json['error_message']['faultstring'])



#
# Copyright 2023 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log

from observabilityclient import client

from aodh.evaluator import threshold
from aodh import keystone_client


LOG = log.getLogger(__name__)
OPTS = [
    cfg.BoolOpt('prometheus_disable_rbac',
                default=False,
                help='Disable RBAC for Prometheus evaluator.',
                deprecated_for_removal=True,
                deprecated_reason="Prometheus RBAC is always disabled. "
                                  "It's not possible to correctly use "
                                  "client-side rbac enforcement from within "
                                  "services. Using it can cause issues.",
                deprecated_since="Flamingo")
]


class PrometheusBase(threshold.ThresholdEvaluator):
    def __init__(self, conf):
        super().__init__(conf)
        self._set_obsclient(conf)

    def _set_obsclient(self, conf):
        session = keystone_client.get_session(conf)
        opts = {'interface': conf.service_credentials.interface,
                'region_name': conf.service_credentials.region_name}
        self._prom = client.Client('1', session, adapter_options=opts)

    def _get_metric_data(self, query):
        LOG.debug(f'Querying Prometheus instance on: {query}')
        return self._prom.query.query(query, disable_rbac=True)


class PrometheusEvaluator(PrometheusBase):

    def _sanitize(self, metric_data):
        sanitized = [float(m.value) for m in metric_data]
        LOG.debug(f'Sanited Prometheus metric data: {metric_data}'
                  f' to statistics: {sanitized}')
        return sanitized

    def evaluate_rule(self, alarm_rule):
        """Evaluate alarm rule.

        :returns: state, trending state, statistics, number of samples outside
        threshold and reason
        """
        metrics = self._get_metric_data(alarm_rule['query'])
        if not metrics:
            LOG.warning("Empty result fetched from Prometheus for query"
                        f" {alarm_rule['query']}")

        statistics = self._sanitize(metrics)
        if not statistics:
            raise threshold.InsufficientDataError('datapoints are unknown',
                                                  statistics)
        return self._process_statistics(alarm_rule, statistics)

    def _unknown_reason_data(self, alarm, statistics):
        LOG.warning(f'Transfering alarm {alarm} on unknown reason')
        last = None if not statistics else statistics[-1]
        return self._reason_data('unknown', len(statistics), last)



# Copyright 2025 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Add cascade

Revision ID: 010
Revises: 009
Create Date: 2025-07-19 12:09:36.651705

"""

# revision identifiers, used by Alembic.
revision = '010'
down_revision = '009'
branch_labels = None
depends_on = None

from alembic import op
from sqlalchemy.engine import reflection


def upgrade():
    inspector = reflection.Inspector.from_engine(op.get_bind())

    fks_to_cascade = {
        'alarm_counter': 'alarm_id',
        'alarm_history': 'alarm_id'
    }

    for table, column in fks_to_cascade.items():
        fk_constraints = inspector.get_foreign_keys(table)
        for fk in fk_constraints:
            if column in fk['constrained_columns']:
                op.drop_constraint(
                    constraint_name=fk['name'],
                    table_name=table,
                    type_='foreignkey'
                )
                op.create_foreign_key(
                    constraint_name=fk['name'],
                    source_table=table,
                    referent_table=fk['referred_table'],
                    local_cols=fk['constrained_columns'],
                    remote_cols=fk['referred_columns'],
                    ondelete='CASCADE'
                )



#
# Copyright 2012 New Dream Network (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Test base classes.
"""

import fixtures
import functools
import os.path
import unittest
import warnings

import oslo_messaging.conffixture
from oslo_utils import timeutils
from oslotest import base
from sqlalchemy import exc as sqla_exc
import webtest

import aodh
from aodh import messaging


class WarningsFixture(fixtures.Fixture):
    """Filters out warnings during test runs."""

    def setUp(self):
        super().setUp()

        self._original_warning_filters = warnings.filters[:]

        warnings.simplefilter('once', DeprecationWarning)

        # FIXME(stephenfin): Determine if we need to replace use of best_match
        warnings.filterwarnings(
            'ignore',
            module='webob',
            message='The behavior of AcceptValidHeader.best_match is ',
            category=DeprecationWarning,
        )

        # FIXME(stephenfin): Determine if we need to replace use of best_match
        warnings.filterwarnings(
            'ignore',
            module='webob',
            message='The behavior of .best_match for the Accept classes is ',
            category=DeprecationWarning,
        )

        # FIXME(stephenfin): Update tests to resolve these issues
        warnings.filterwarnings(
            'ignore',
            module='oslo_policy',
            message='Policy ".*": ".*" failed scope check. ',
            category=UserWarning,
        )

        # Enable deprecation warnings for aodh itself to capture upcoming
        # SQLAlchemy changes

        warnings.filterwarnings(
            'ignore',
            category=sqla_exc.SADeprecationWarning,
        )

        warnings.filterwarnings(
            'error',
            module='aodh',
            category=sqla_exc.SADeprecationWarning,
        )

        # Enable general SQLAlchemy warnings also to ensure we're not doing
        # silly stuff. It's possible that we'll need to filter things out here
        # with future SQLAlchemy versions, but that's a good thing

        warnings.filterwarnings(
            'error',
            module='aodh',
            category=sqla_exc.SAWarning,
        )

        self.addCleanup(self._reset_warning_filters)

    def _reset_warning_filters(self):
        warnings.filters[:] = self._original_warning_filters


class BaseTestCase(base.BaseTestCase):
    def setup_messaging(self, conf, exchange=None):
        self.useFixture(oslo_messaging.conffixture.ConfFixture(conf))
        self.useFixture(WarningsFixture())
        conf.set_override("notification_driver", ["messaging"])
        if not exchange:
            exchange = 'aodh'
        conf.set_override("control_exchange", exchange)

        # NOTE(sileht): Ensure a new oslo.messaging driver is loaded
        # between each tests
        self.transport = messaging.get_transport(conf, "fake://", cache=False)
        self.useFixture(fixtures.MockPatch(
            'aodh.messaging.get_transport',
            return_value=self.transport))

    def assertTimestampEqual(self, first, second, msg=None):
        """Checks that two timestamps are equals.

        This relies on assertAlmostEqual to avoid rounding problem, and only
        checks up the first microsecond values.

        """
        return self.assertAlmostEqual(
            timeutils.delta_seconds(first, second),
            0.0,
            places=5)

    @staticmethod
    def path_get(project_file=None):
        root = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                            '..',
                                            '..',
                                            )
                               )
        if project_file:
            return os.path.join(root, project_file)
        else:
            return root

    def assert_single_item(self, items, **filters):
        return self.assert_multiple_items(items, 1, **filters)[0]

    def assert_multiple_items(self, items, count, **filters):
        def _matches(item, **props):
            for prop_name, prop_val in props.items():
                v = (item[prop_name] if isinstance(item, dict)
                     else getattr(item, prop_name))
                if v != prop_val:
                    return False
            return True

        filtered_items = list(
            [item for item in items if _matches(item, **filters)]
        )
        found = len(filtered_items)

        if found != count:
            self.fail("Wrong number of items found [filters=%s, "
                      "expected=%s, found=%s]" % (filters, count, found))

        return filtered_items


def _skip_decorator(func):
    @functools.wraps(func)
    def skip_if_not_implemented(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except aodh.NotImplementedError as e:
            raise unittest.SkipTest(str(e))
        except webtest.app.AppError as e:
            if 'not implemented' in str(e):
                raise unittest.SkipTest(str(e))
            raise
    return skip_if_not_implemented


class SkipNotImplementedMeta(type):
    def __new__(cls, name, bases, local):
        for attr in local:
            value = local[attr]
            if callable(value) and (
                    attr.startswith('test_') or attr == 'setUp'):
                local[attr] = _skip_decorator(value)
        return type.__new__(cls, name, bases, local)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""SQLAlchemy storage backend."""

import copy
import datetime
import os.path
import threading

from alembic import command
from alembic import config
from alembic import migration
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import utils as oslo_sql_utils
from oslo_log import log
from oslo_utils import importutils
from oslo_utils import timeutils
import sqlalchemy
from sqlalchemy import asc
from sqlalchemy import desc
from sqlalchemy.engine import url as sqlalchemy_url
from sqlalchemy import func
from sqlalchemy.orm import exc

import aodh
from aodh import storage
from aodh.storage import base
from aodh.storage import models as alarm_api_models
from aodh.storage.sqlalchemy import models
from aodh.storage.sqlalchemy import utils as sql_utils

DB_CONFIGURED = False

osprofiler_sqlalchemy = importutils.try_import('osprofiler.sqlalchemy')

_CONTEXT = threading.local()
LOG = log.getLogger(__name__)

AVAILABLE_CAPABILITIES = {
    'alarms': {'query': {'simple': True,
                         'complex': True},
               'history': {'query': {'simple': True,
                                     'complex': True}}},
}
AVAILABLE_STORAGE_CAPABILITIES = {
    'storage': {'production_ready': True},
}

# int type should be 32b long in both mysql and postgresql
COUNTER_RESET_AT_VALUE = 2000000000


def apply_filters(query, model, **filters):
    filter_dict = {}

    for key, value in filters.items():
        column_attr = getattr(model, key)

        if isinstance(value, dict):
            if 'in' in value:
                query = query.filter(column_attr.in_(value['in']))
            elif 'nin' in value:
                query = query.filter(~column_attr.in_(value['nin']))
            elif 'ne' in value:
                query = query.filter(column_attr != value['ne'])
            elif 'gt' in value:
                query = query.filter(column_attr > value['gt'])
            elif 'ge' in value:
                query = query.filter(column_attr >= value['ge'])
            elif 'lt' in value:
                query = query.filter(column_attr < value['lt'])
            elif 'le' in value:
                query = query.filter(column_attr <= value['le'])
            elif 'eq' in value:
                query = query.filter(column_attr == value['eq'])
            elif 'has' in value:
                like_pattern = '%{}%'.format(value['has'])
                query = query.filter(column_attr.like(like_pattern))
        else:
            filter_dict[key] = value

    if filter_dict:
        query = query.filter_by(**filter_dict)

    return query


def _session_for_read():
    session = enginefacade.reader.using(_CONTEXT)
    if osprofiler_sqlalchemy:
        session = osprofiler_sqlalchemy.wrap_session(sqlalchemy, session)
    return session


def _session_for_write():
    session = enginefacade.writer.using(_CONTEXT)
    if osprofiler_sqlalchemy:
        session = osprofiler_sqlalchemy.wrap_session(sqlalchemy, session)
    return session


class Connection(base.Connection):
    """Put the data into a SQLAlchemy database. """
    CAPABILITIES = base.update_nested(base.Connection.CAPABILITIES,
                                      AVAILABLE_CAPABILITIES)
    STORAGE_CAPABILITIES = base.update_nested(
        base.Connection.STORAGE_CAPABILITIES,
        AVAILABLE_STORAGE_CAPABILITIES,
    )

    def __init__(self, conf):
        # Set max_retries to 0, since oslo.db in certain cases may attempt
        # to retry making the db connection retried max_retries ^ 2 times
        # in failure case and db reconnection has already been implemented
        # in storage.__init__.get_connection_from_config function
        global DB_CONFIGURED

        if not DB_CONFIGURED:
            options = dict(conf.database.items())
            options['connection'] = self.dress_url(conf.database.connection)
            options['max_retries'] = 0
            options['sqlite_fk'] = True
            # oslo.db doesn't support options defined by Aodh
            for opt in storage.OPTS:
                options.pop(opt.name, None)

            enginefacade.configure(**options)

            DB_CONFIGURED = True

        self.conf = conf

    @staticmethod
    def dress_url(url):
        # If no explicit driver has been set, we default to pymysql
        if url.startswith("mysql://"):
            url = sqlalchemy_url.make_url(url)
            url = url.set(drivername="mysql+pymysql")
            return str(url)
        return url

    def _get_alembic_config(self):
        path = os.path.join(os.path.dirname(__file__),
                            "sqlalchemy/alembic/alembic.ini")
        cfg = config.Config(os.path.abspath(path))
        cfg.attributes['configure_logger'] = False
        return cfg

    def upgrade(self, nocreate=False):
        cfg = self._get_alembic_config()

        engine = enginefacade.writer.get_engine()
        engine_url = engine.url.render_as_string(
            hide_password=False).replace('%', '%%')
        cfg.set_main_option('sqlalchemy.url', engine_url)

        with engine.connect() as conn, conn.begin():
            cfg.attributes['connection'] = conn
            if nocreate:
                command.upgrade(cfg, "head")
            else:
                ctxt = migration.MigrationContext.configure(conn)
                current_version = ctxt.get_current_revision()
                if current_version is None:
                    models.Base.metadata.create_all(conn, checkfirst=False)
                    command.stamp(cfg, "head")
                else:
                    command.upgrade(cfg, "head")

    def clear(self):
        engine = enginefacade.writer.get_engine()
        for table in reversed(models.Base.metadata.sorted_tables):
            with engine.connect() as conn, conn.begin():
                conn.execute(table.delete())
        engine.dispose()

    def _retrieve_data(self, filter_expr, orderby, limit, table):
        if limit == 0:
            return []

        engine = enginefacade.reader.get_engine()
        with _session_for_read() as session:
            query = session.query(table)

        transformer = sql_utils.QueryTransformer(table, query,
                                                 dialect=engine.dialect.name)
        if filter_expr is not None:
            transformer.apply_filter(filter_expr)

        transformer.apply_options(orderby,
                                  limit)

        retrieve = {models.Alarm: self._retrieve_alarms,
                    models.AlarmChange: self._retrieve_alarm_history}
        return retrieve[table](transformer.get_query())

    @staticmethod
    def _row_to_alarm_model(row):
        return alarm_api_models.Alarm(
            alarm_id=row.alarm_id,
            enabled=row.enabled,
            type=row.type,
            name=row.name,
            description=row.description,
            timestamp=row.timestamp,
            user_id=row.user_id,
            project_id=row.project_id,
            state=row.state,
            state_timestamp=row.state_timestamp,
            state_reason=row.state_reason,
            ok_actions=row.ok_actions,
            alarm_actions=row.alarm_actions,
            insufficient_data_actions=(row.insufficient_data_actions),
            rule=row.rule,
            time_constraints=row.time_constraints,
            repeat_actions=row.repeat_actions,
            severity=row.severity,
            evaluate_timestamp=row.evaluate_timestamp
        )

    def _retrieve_alarms(self, query):
        return [self._row_to_alarm_model(x) for x in query.all()]

    @staticmethod
    def _get_pagination_query(session, query, pagination, api_model, model):
        if not pagination.get('sort'):
            pagination['sort'] = api_model.DEFAULT_SORT
        marker = None
        if pagination.get('marker'):
            key_attr = getattr(model, api_model.PRIMARY_KEY)
            marker_query = copy.copy(query)
            marker_query = marker_query.filter(
                key_attr == pagination['marker'])
            try:
                marker = marker_query.limit(1).one()
            except exc.NoResultFound:
                raise storage.InvalidMarker(
                    'Marker %s not found.' % pagination['marker'])
        limit = pagination.get('limit')
        # we sort by "severity" by its semantic than its alphabetical
        # order when "severity" specified in sorts.
        for sort_key, sort_dir in pagination['sort'][::-1]:
            if sort_key == 'severity':
                engine = session.connection()
                if engine.dialect.name != "mysql":
                    raise aodh.NotImplementedError
                sort_dir_func = {'asc': asc, 'desc': desc}[sort_dir]
                query = query.order_by(sort_dir_func(
                    func.field(getattr(model, sort_key), 'low',
                               'moderate', 'critical')))
                pagination['sort'].remove((sort_key, sort_dir))

        sort_keys = [s[0] for s in pagination['sort']]
        sort_dirs = [s[1] for s in pagination['sort']]
        return oslo_sql_utils.paginate_query(
            query, model, limit, sort_keys, sort_dirs=sort_dirs, marker=marker)

    def get_alarms(self, meter=None, pagination=None, **kwargs):
        """Yields a lists of alarms that match filters."""
        pagination = pagination or {}
        with _session_for_read() as session:
            query = session.query(models.Alarm)
            query = apply_filters(query, models.Alarm, **kwargs)
            query = self._get_pagination_query(
                session,
                query,
                pagination,
                alarm_api_models.Alarm,
                models.Alarm,
            )
            alarms = [self._row_to_alarm_model(x) for x in query.all()]

        # TODO(cmart): improve this by using sqlalchemy.func factory
        if meter is not None:
            alarms = filter(lambda row:
                            row.rule.get('meter_name', None) == meter,
                            alarms)

        return alarms

    def create_alarm(self, alarm):
        """Create an alarm.

        :param alarm: The alarm to create.
        """
        with _session_for_write() as session:
            alarm_row = models.Alarm(alarm_id=alarm.alarm_id)
            alarm_row.update(alarm.as_dict())
            session.add(alarm_row)

        return self._row_to_alarm_model(alarm_row)

    def update_alarm(self, alarm):
        """Update an alarm.

        :param alarm: the new Alarm to update
        """
        with _session_for_write() as session:
            count = session.query(models.Alarm).filter(
                models.Alarm.alarm_id == alarm.alarm_id,
            ).update(alarm.as_dict())

        if not count:
            raise storage.AlarmNotFound(alarm.alarm_id)

        return alarm

    def delete_alarm(self, alarm_id):
        """Delete an alarm and its history data.

        :param alarm_id: ID of the alarm to delete
        """
        with _session_for_write() as session:
            session.query(models.Alarm).filter(
                models.Alarm.alarm_id == alarm_id,
            ).delete()

    def increment_alarm_counter(self, alarm_id, project_id, state):
        """Increment a counter.

        :param alarm_id: the id of alarm to which the counter belongs
        :param project_id: the id of the project of the alarm
        :param state: the state of the alarm to increment
        """
        with _session_for_write() as session:
            LOG.debug("Incrementing counter %(state)s for alarm %(alarm_id)s",
                      {'alarm_id': alarm_id, 'state': state})

            counter = self.get_alarm_counters(alarm_id, project_id, state)
            counter_value = 0
            if len(counter) == 1:
                counter_value = counter[0].value
            new_value = counter_value + 1
            if counter_value >= COUNTER_RESET_AT_VALUE:
                LOG.debug("Resetting counter %(state)s for alarm %(alarm_id)s",
                          {'alarm_id': alarm_id, 'state': state})
                new_value = 1

            elif counter_value == 0:
                # We have a new uninitialized counter
                counter = models.AlarmCounter(
                    alarm_id=alarm_id,
                    project_id=project_id,
                    state=state
                )
                counter.update({'value': new_value})
                session.add(counter)
            else:
                session.query(models.AlarmCounter).filter(
                    models.AlarmCounter.alarm_id == alarm_id,
                    models.AlarmCounter.project_id == project_id,
                    models.AlarmCounter.state == state,
                ).update({'value': new_value})

        return state

    @staticmethod
    def _row_to_alarm_change_model(row):
        return alarm_api_models.AlarmChange(event_id=row.event_id,
                                            alarm_id=row.alarm_id,
                                            type=row.type,
                                            detail=row.detail,
                                            user_id=row.user_id,
                                            project_id=row.project_id,
                                            on_behalf_of=row.on_behalf_of,
                                            timestamp=row.timestamp,
                                            severity=row.severity)

    def query_alarms(self, filter_expr=None, orderby=None, limit=None):
        """Yields a lists of alarms that match filter."""
        return self._retrieve_data(filter_expr, orderby, limit, models.Alarm)

    def get_alarm_counters(self, alarm_id=None, project_id=None, state=None):
        """Yields a counter based on its alarm_id, project_id and state."""
        filters = {}
        if alarm_id is not None:
            filters['alarm_id'] = alarm_id
        if project_id is not None:
            filters['project_id'] = project_id
        if state is not None:
            filters['state'] = state
        with _session_for_read() as session:
            counters = session.query(models.AlarmCounter).filter_by(
                **filters
            ).all()
            if counters is None:
                return []
        return counters

    def _retrieve_alarm_history(self, query):
        return (self._row_to_alarm_change_model(x) for x in query.all())

    def query_alarm_history(self, filter_expr=None, orderby=None, limit=None):
        """Return an iterable of model.AlarmChange objects."""
        return self._retrieve_data(filter_expr,
                                   orderby,
                                   limit,
                                   models.AlarmChange)

    def get_alarm_changes(self, alarm_id, on_behalf_of,
                          user=None, project=None, alarm_type=None,
                          severity=None, start_timestamp=None,
                          start_timestamp_op=None, end_timestamp=None,
                          end_timestamp_op=None, pagination=None):
        """Yields list of AlarmChanges describing alarm history

        Changes are always sorted in reverse order of occurrence, given
        the importance of currency.

        Segregation for non-administrative users is done on the basis
        of the on_behalf_of parameter. This allows such users to have
        visibility on both the changes initiated by themselves directly
        (generally creation, rule changes, or deletion) and also on those
        changes initiated on their behalf by the alarming service (state
        transitions after alarm thresholds are crossed).

        :param alarm_id: ID of alarm to return changes for
        :param on_behalf_of: ID of tenant to scope changes query (None for
                             administrative user, indicating all projects)
        :param user: Optional ID of user to return changes for
        :param project: Optional ID of project to return changes for
        :param alarm_type: Optional change type
        :param severity: Optional alarm severity
        :param start_timestamp: Optional modified timestamp start range
        :param start_timestamp_op: Optional timestamp start range operation
        :param end_timestamp: Optional modified timestamp end range
        :param end_timestamp_op: Optional timestamp end range operation
        :param pagination: Pagination query parameters.
        """
        pagination = pagination or {}

        with _session_for_read() as session:
            query = session.query(models.AlarmChange)
            query = query.filter(models.AlarmChange.alarm_id == alarm_id)

            if on_behalf_of is not None:
                query = query.filter(
                    models.AlarmChange.on_behalf_of == on_behalf_of)
            if user is not None:
                query = query.filter(models.AlarmChange.user_id == user)
            if project is not None:
                query = query.filter(models.AlarmChange.project_id == project)
            if alarm_type is not None:
                query = query.filter(models.AlarmChange.type == alarm_type)
            if severity is not None:
                query = query.filter(models.AlarmChange.severity == severity)
            if start_timestamp:
                if start_timestamp_op == 'gt':
                    query = query.filter(
                        models.AlarmChange.timestamp > start_timestamp)
                else:
                    query = query.filter(
                        models.AlarmChange.timestamp >= start_timestamp)
            if end_timestamp:
                if end_timestamp_op == 'le':
                    query = query.filter(
                        models.AlarmChange.timestamp <= end_timestamp)
                else:
                    query = query.filter(
                        models.AlarmChange.timestamp < end_timestamp)

            query = self._get_pagination_query(
                session, query, pagination, alarm_api_models.AlarmChange,
                models.AlarmChange)

            return (self._row_to_alarm_change_model(x) for x in query.all())

    def record_alarm_change(self, alarm_change):
        """Record alarm change event."""
        with _session_for_write() as session:
            alarm_change_row = models.AlarmChange(
                event_id=alarm_change['event_id'])
            alarm_change_row.update(alarm_change)
            session.add(alarm_change_row)

    def clear_expired_alarm_history_data(self, ttl, max_count=100):
        """Clear expired alarm history data from the backend storage system.

        Clearing occurs according to the time-to-live.

        :param ttl: Number of seconds to keep alarm history records for.
        :param max_count: Number of records to delete.
        """
        with _session_for_write() as session:
            end = timeutils.utcnow() - datetime.timedelta(seconds=ttl)
            alarm_history_q = (session.query(models.AlarmChange.event_id)
                               .filter(models.AlarmChange.timestamp < end))
            event_ids = [i[0] for i in alarm_history_q.limit(max_count)]
            deleted_rows = session.query(models.AlarmChange).filter(
                models.AlarmChange.event_id.in_(event_ids)
            ).delete(synchronize_session="fetch")

        LOG.info("%d alarm histories are removed from database", deleted_rows)

    def conditional_update(self, model, values, expected_values, filters=None):
        """Compare-and-swap conditional update SQLAlchemy implementation."""
        filters = filters or {}
        filters.update(expected_values)

        with _session_for_write() as session:
            query = session.query(model)
            if filters:
                query = query.filter_by(**filters)

            update_args = {'synchronize_session': False}

            result = query.update(values, **update_args)

        return 0 != result

    @staticmethod
    def _row_to_quota_model(row):
        return alarm_api_models.Quota(
            project_id=row.project_id,
            resource=row.resource,
            limit=row.limit,
        )

    def get_quotas(self, project_id):
        """Get resource quota for the given project."""
        filters = {'project_id': project_id}
        with _session_for_read() as session:
            query = session.query(models.Quota).filter_by(**filters)
            return [self._row_to_quota_model(x) for x in query.all()]

    def set_quotas(self, project_id, quotas):
        """Set resource quota for the given user."""
        with _session_for_write() as session:
            for q in quotas:
                values = {
                    'project_id': project_id,
                    'resource': q['resource'],
                }

                quota = session.query(models.Quota).filter_by(**values).first()
                if not quota:
                    new_quota = models.Quota(project_id=project_id,
                                             resource=q['resource'],
                                             limit=q['limit'])
                    session.add(new_quota)
                else:
                    values['limit'] = q['limit']
                    quota.update(values.copy())

            filters = {'project_id': project_id}
            query = session.query(models.Quota).filter_by(**filters)
            return [self._row_to_quota_model(x) for x in query.all()]

    def delete_quotas(self, project_id):
        filters = {'project_id': project_id}
        with _session_for_write() as session:
            session.query(models.Quota).filter_by(**filters).delete()



#
# Copyright 2013-2015 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import operator

from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils

from aodh import evaluator

LOG = log.getLogger(__name__)

COMPARATORS = {
    'gt': operator.gt,
    'lt': operator.lt,
    'ge': operator.ge,
    'le': operator.le,
    'eq': operator.eq,
    'ne': operator.ne,
}

OPTS = [
    cfg.IntOpt('additional_ingestion_lag',
               min=0,
               max=3600 * 24,
               default=0,
               help='The number of seconds to extend the evaluation windows '
               'to compensate the reporting/ingestion lag.')
]


class InsufficientDataError(Exception):
    def __init__(self, reason, statistics):
        self.reason = reason
        self.statistics = statistics
        super().__init__(reason)


class ThresholdEvaluator(evaluator.Evaluator):

    # the sliding evaluation window is extended to allow
    # the reporting/ingestion lag this can be increased
    # with 'additional_ingestion_lag' seconds if needed.
    look_back = 1

    def _bound_duration(self, rule):
        """Bound the duration of the statistics query."""
        now = timeutils.utcnow()
        # when exclusion of weak datapoints is enabled, we extend
        # the look-back period so as to allow a clearer sample count
        # trend to be established
        window = ((rule.get('period', None) or rule['granularity'])
                  * (rule['evaluation_periods'] + self.look_back) +
                  self.conf.additional_ingestion_lag)
        start = now - datetime.timedelta(seconds=window)
        LOG.debug('query stats from %(start)s to '
                  '%(now)s', {'start': start, 'now': now})
        return start.isoformat(), now.isoformat()

    @staticmethod
    def _reason_data(disposition, count, most_recent):
        """Create a reason data dictionary for this evaluator type."""
        return {'type': 'threshold', 'disposition': disposition,
                'count': count, 'most_recent': most_recent}

    @classmethod
    def _reason(cls, alarm, statistics, state, count):
        """Fabricate reason string."""
        if state == evaluator.OK:
            disposition = 'inside'
            count = len(statistics) - count
        else:
            disposition = 'outside'
        last = statistics[-1] if statistics else None
        transition = alarm.state != state
        reason_data = cls._reason_data(disposition, count, last)
        if transition:
            return ('Transition to %(state)s due to %(count)d samples'
                    ' %(disposition)s threshold, most recent:'
                    ' %(most_recent)s' % dict(reason_data, state=state),
                    reason_data)
        return ('Remaining as %(state)s due to %(count)d samples'
                ' %(disposition)s threshold, most recent: %(most_recent)s'
                % dict(reason_data, state=state), reason_data)

    def _process_statistics(self, alarm_rule, statistics):

        def _compare(value):
            op = COMPARATORS[alarm_rule['comparison_operator']]
            limit = alarm_rule['threshold']
            LOG.debug('comparing value %(value)s against threshold'
                      ' %(limit)s', {'value': value, 'limit': limit})
            return op(value, limit)

        compared = list(map(_compare, statistics))
        distilled = all(compared)
        unequivocal = distilled or not any(compared)
        number_outside = len([c for c in compared if c])

        if unequivocal:
            state = evaluator.ALARM if distilled else evaluator.OK
            return state, None, statistics, number_outside, None
        else:
            trending_state = evaluator.ALARM if compared[-1] else evaluator.OK
            return None, trending_state, statistics, number_outside, None

    def evaluate_rule(self, alarm_rule):
        """Evaluate alarm rule.

        :returns: state, trending state and statistics.
        """
        start, end = self._bound_duration(alarm_rule)
        statistics = self._statistics(alarm_rule, start, end)
        statistics = self._sanitize(alarm_rule, statistics)
        sufficient = len(statistics) >= alarm_rule['evaluation_periods']
        if not sufficient:
            raise InsufficientDataError(
                '%d datapoints are unknown' % alarm_rule['evaluation_periods'],
                statistics)

        return self._process_statistics(alarm_rule, statistics)

    def _unknown_reason_data(self, alarm, statistics):
        LOG.warning(f'Expecting {alarm.rule["evaluation_periods"]} datapoints'
                    f' but only get {len(statistics)}')
        # Reason is not same as log message because we want to keep
        # consistent since thirdparty software may depend on old format.
        last = None if not statistics else statistics[-1]
        return self._reason_data('unknown', alarm.rule['evaluation_periods'],
                                 last)

    def _transition_alarm(self, alarm, state, trending_state, statistics,
                          outside_count, unknown_reason):
        unknown = alarm.state == evaluator.UNKNOWN
        continuous = alarm.repeat_actions

        if trending_state:
            if unknown or continuous:
                state = trending_state if unknown else alarm.state
                reason, reason_data = self._reason(alarm, statistics, state,
                                                   outside_count)
                self._refresh(alarm, state, reason, reason_data)
                return

        if state == evaluator.UNKNOWN and not unknown:
            reason_data = self._unknown_reason_data(alarm, statistics)
            self._refresh(alarm, state, unknown_reason, reason_data)

        elif state and (alarm.state != state or continuous):
            reason, reason_data = self._reason(alarm, statistics, state,
                                               outside_count)
            self._refresh(alarm, state, reason, reason_data)

    def evaluate(self, alarm):
        if not self.within_time_constraint(alarm):
            LOG.debug('Attempted to evaluate alarm %s, but it is not '
                      'within its time constraint.', alarm.alarm_id)
            return

        try:
            evaluation = self.evaluate_rule(alarm.rule)
        except InsufficientDataError as e:
            evaluation = (evaluator.UNKNOWN, None, e.statistics, 0,
                          e.reason)
        self._transition_alarm(alarm, *evaluation)
        if evaluation[0] is not None:
            self._increment_evaluation_result(alarm.alarm_id,
                                              alarm.project_id,
                                              evaluation[0].replace(" ", "_"))



#
# Copyright 2012 New Dream Network, LLC (DreamHost)
# Copyright 2014 Hewlett-Packard Company
# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Access Control Lists (ACL's) control access the API server."""

from oslo_context import context
import pecan


def target_from_segregation_rule(req, enforcer):
    """Return a target corresponding to an alarm returned by segregation rule

    This allows to use project_id: in an oslo_policy rule for query/listing.

    :param req: Webob Request object
    :param enforcer: policy enforcer

    :returns: target
    """

    project_id = get_limited_to_project(req, enforcer)
    if project_id is not None:
        return {'project_id': project_id}
    return {}


def enforce(policy_name, req, enforcer, target):
    """Return the user and project the request should be limited to.

    :param policy_name: the policy name to validate authz against.
    :param req: Webob Request object
    :param enforcer: policy enforcer
    :param target: the alarm or "auto" to

    """
    rule_method = "telemetry:" + policy_name
    ctxt = context.RequestContext.from_environ(req.environ)

    if not enforcer.enforce(rule_method, target, ctxt.to_dict()):
        pecan.core.abort(status_code=403,
                         detail='RBAC Authorization Failed')


# TODO(fabiog): these methods are still used because the scoping part is really
# convoluted and difficult to separate out.

def get_limited_to(req, enforcer):
    """Return the user and project the request should be limited to.

    :param req: Webob Request object
    :param enforcer: policy enforcer
    :return: A tuple of (user, project), set to None if there's no limit on
    one of these.

    """
    ctxt = context.RequestContext.from_environ(req.environ)
    if not enforcer.enforce('segregation', {}, ctxt.to_dict()):
        return ctxt.user_id, ctxt.project_id

    return None, None


def get_limited_to_project(req, enforcer):
    """Return the project the request should be limited to.

    :param req: Webob Request object
    :param enforcer: policy enforcer
    :return: A project, or None if there's no limit on it.

    """
    return get_limited_to(req, enforcer)[1]


def is_admin(req, enforcer):
    ctxt = context.RequestContext.from_environ(req.environ)
    return enforcer.enforce('context_is_admin', {}, ctxt.to_dict())



#
# Copyright 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log

from aodh import service
from aodh import storage


LOG = log.getLogger(__name__)


def dbsync():
    conf = service.prepare_service()
    storage.get_connection_from_config(conf).upgrade()


def expirer():
    conf = service.prepare_service()

    if conf.database.alarm_history_time_to_live > 0:
        LOG.debug("Clearing expired alarm history data")
        conn = storage.get_connection_from_config(conf)
        max_count = conf.database.alarm_histories_delete_batch_size
        try:
            if max_count > 0:
                conn.clear_expired_alarm_history_data(
                    conf.database.alarm_history_time_to_live,
                    max_count)
            else:
                deleted = max_count = 100
                while deleted and deleted > 0:
                    deleted = conn.clear_expired_alarm_history_data(
                        conf.database.alarm_history_time_to_live,
                        max_count)
        except TypeError:
            LOG.warning("Storage driver does not support "
                        "'alarm_histories_delete_batch_size' config option.")
    else:
        LOG.info("Nothing to clean, database alarm history time to live "
                 "is disabled")



# Copyright 2016 Hewlett Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_middleware import cors
from oslo_policy import opts as policy_opts


def set_lib_defaults():
    """Update default value for configuration options from other namespace.

    Example, oslo lib config options. This is needed for
    config generator tool to pick these default value changes.
    https://docs.openstack.org/oslo.config/latest/cli/
    generator.html#modifying-defaults-from-other-namespaces
    """
    set_cors_middleware_defaults()

    # Update default value of oslo.policy policy_file, ,
    # enforce_scope, and enforce_new_defaults config options.
    policy_opts.set_defaults(cfg.CONF, 'policy.yaml',
                             enforce_scope=False,
                             enforce_new_defaults=False)


def set_cors_middleware_defaults():
    """Update default configuration options for oslo.middleware."""
    cors.set_defaults(
        allow_headers=['X-Auth-Token',
                       'X-Openstack-Request-Id',
                       'X-Subject-Token'],
        expose_headers=['X-Auth-Token',
                        'X-Openstack-Request-Id',
                        'X-Subject-Token'],
        allow_methods=['GET',
                       'PUT',
                       'POST',
                       'DELETE',
                       'PATCH']
    )



#
# Copyright 2014 OpenStack Foundation
# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import cotyledon
from cotyledon import oslo_config_glue
from oslo_log import log

from aodh import evaluator as evaluator_svc
from aodh import event as event_svc
from aodh import notifier as notifier_svc
from aodh import service

LOG = log.getLogger(__name__)


def notifier():
    sm = cotyledon.ServiceManager()
    conf = service.prepare_service()
    oslo_config_glue.setup(sm, conf)
    sm.add(notifier_svc.AlarmNotifierService,
           workers=conf.notifier.workers, args=(conf,))
    sm.run()


def evaluator():
    sm = cotyledon.ServiceManager()
    conf = service.prepare_service()
    oslo_config_glue.setup(sm, conf)
    sm.add(evaluator_svc.AlarmEvaluationService,
           workers=conf.evaluator.workers, args=(conf,))
    sm.run()


def listener():
    sm = cotyledon.ServiceManager()
    conf = service.prepare_service()
    oslo_config_glue.setup(sm, conf)
    sm.add(event_svc.EventAlarmEvaluationService,
           workers=conf.listener.workers, args=(conf,))
    sm.run()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
SQLAlchemy models for aodh data.
"""
import json

from oslo_utils import timeutils
from oslo_utils import uuidutils
import sqlalchemy as sa
from sqlalchemy import Column, String, Index, Boolean, Text, DateTime, Integer
from sqlalchemy.dialects import mysql
from sqlalchemy.orm import declarative_base
from sqlalchemy.types import TypeDecorator


class JSONEncodedDict(TypeDecorator):
    """Represents an immutable structure as a json-encoded string."""

    impl = Text
    cache_ok = True

    @staticmethod
    def process_bind_param(value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    @staticmethod
    def process_result_value(value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class TimestampUTC(TypeDecorator):
    """Represents a timestamp precise to the microsecond."""

    impl = DateTime
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == 'mysql':
            return dialect.type_descriptor(mysql.DATETIME(fsp=6))
        return self.impl


class AodhBase:
    """Base class for Aodh Models."""
    __table_args__ = {'mysql_charset': "utf8",
                      'mysql_engine': "InnoDB"}
    __table_initialized__ = False

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def update(self, values):
        """Make the model object behave like a dict."""
        for k, v in values.items():
            setattr(self, k, v)


Base = declarative_base(cls=AodhBase)


class Alarm(Base):
    """Define Alarm data."""
    __tablename__ = 'alarm'
    __table_args__ = (
        Index('ix_alarm_user_id', 'user_id'),
        Index('ix_alarm_project_id', 'project_id'),
        Index('ix_alarm_enabled', 'enabled'),
        Index('ix_alarm_type', 'type'),
    )
    alarm_id = Column(String(128), primary_key=True)
    enabled = Column(Boolean)
    name = Column(Text)
    type = Column(String(50))
    severity = Column(String(50))
    description = Column(Text)
    timestamp = Column(TimestampUTC, default=lambda: timeutils.utcnow())

    user_id = Column(String(128))
    project_id = Column(String(128))

    state = Column(String(255))
    state_reason = Column(Text)
    state_timestamp = Column(TimestampUTC,
                             default=lambda: timeutils.utcnow())

    ok_actions = Column(JSONEncodedDict)
    alarm_actions = Column(JSONEncodedDict)
    insufficient_data_actions = Column(JSONEncodedDict)
    repeat_actions = Column(Boolean)

    rule = Column(JSONEncodedDict)
    time_constraints = Column(JSONEncodedDict)

    evaluate_timestamp = Column(DateTime, default=lambda: timeutils.utcnow())


class AlarmChange(Base):
    """Define AlarmChange data."""
    __tablename__ = 'alarm_history'
    __table_args__ = (
        Index('ix_alarm_history_alarm_id', 'alarm_id'),
    )
    event_id = Column(String(128), primary_key=True)
    alarm_id = Column(String(128),
                      sa.ForeignKey('alarm.alarm_id', ondelete='CASCADE'),
                      nullable=False)
    on_behalf_of = Column(String(128))
    project_id = Column(String(128))
    user_id = Column(String(128))
    type = Column(String(20))
    detail = Column(Text)
    timestamp = Column(TimestampUTC, default=lambda: timeutils.utcnow())
    severity = Column(String(50))


class Quota(Base):
    __tablename__ = 'quota'
    __table_args__ = (
        sa.UniqueConstraint('project_id', 'resource'),
        Index('ix_%s_project_id_resource' % __tablename__,
              'project_id', 'resource'),
    )

    id = Column(String(36), primary_key=True, default=uuidutils.generate_uuid)
    project_id = Column(String(128), nullable=False)
    resource = Column(String(50), nullable=False)
    limit = Column(Integer, nullable=False)


class AlarmCounter(Base):
    __tablename__ = 'alarm_counter'
    __table_args__ = (
        sa.UniqueConstraint('alarm_id', 'project_id', 'state'),
        Index('ix_%s_alarm_id' % __tablename__,
              'alarm_id'),
        Index('ix_%s_state' % __tablename__,
              'state'),
        Index('ix_%s_project_id' % __tablename__,
              'project_id'),
    )

    id = Column(String(36), primary_key=True, default=uuidutils.generate_uuid)
    alarm_id = Column(String(128),
                      sa.ForeignKey('alarm.alarm_id', ondelete='CASCADE'),
                      nullable=False)
    project_id = Column(String(128), nullable=False)
    state = Column(String(128), nullable=False)
    value = Column(Integer, nullable=False)



#
# Copyright 2014 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Rest alarm notifier with trusted authentication."""

from urllib import parse

from aodh import keystone_client
from aodh.notifier import rest


class TrustAlarmNotifierMixin:
    """Mixin class to add Keystone trust support to an AlarmNotifier.

    Provides a notify() method that interprets the trust ID and then calls
    the parent class's notify(), passing the necessary authentication data in
    the headers.
    """

    def notify(self, action, alarm_id, alarm_name, severity, previous, current,
               reason, reason_data):
        trust_id = action.username

        client = keystone_client.get_trusted_client(self.conf, trust_id)

        # Remove the fake user
        netloc = action.netloc.split("@")[1]
        # Remove the trust prefix
        scheme = action.scheme[6:]

        action = parse.SplitResult(scheme, netloc, action.path, action.query,
                                   action.fragment)

        headers = {'X-Auth-Token': keystone_client.get_auth_token(client)}
        super().notify(
            action, alarm_id, alarm_name, severity, previous, current, reason,
            reason_data, headers)


class TrustRestAlarmNotifier(TrustAlarmNotifierMixin, rest.RestAlarmNotifier):
    """Notifier supporting keystone trust authentication.

    This alarm notifier is intended to be used to call an endpoint using
    keystone authentication. It uses the aodh service user to
    authenticate using the trust ID provided.

    The URL must be in the form ``trust+http://host/action``.
    """



# Copyright 2019 Catalyst Cloud Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime

from dateutil import parser
from octaviaclient.api.v2 import octavia
from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils

from aodh import evaluator
from aodh.evaluator import threshold
from aodh import keystone_client as aodh_keystone

LOG = log.getLogger(__name__)

ALARM_TYPE = "loadbalancer_member_health"

OPTS = [
    cfg.IntOpt('member_creation_time',
               default=120,
               help='The time in seconds to wait for the load balancer '
                    'member creation.'
               ),
]


class LoadBalancerMemberHealthEvaluator(evaluator.Evaluator):
    def __init__(self, conf):
        super().__init__(conf)
        self._lb_client = None

    @property
    def lb_client(self):
        if self._lb_client is None:
            endpoint = aodh_keystone.url_for(
                self.conf,
                service_type='load-balancer',
                interface=self.conf.service_credentials.interface,
                region_name=self.conf.service_credentials.region_name
            )
            self._lb_client = octavia.OctaviaAPI(
                session=aodh_keystone.get_session(self.conf),
                service_type='load-balancer',
                endpoint=endpoint
            )

        return self._lb_client

    def _get_unhealthy_members(self, pool_id):
        """Get number of unhealthy members in a pool.

        The member(virutual machine) operating_status keeps ERROR after
        creation before the application is up and running inside, it should be
        ignored during the check.
        """
        unhealthy_members = []

        try:
            ret = self.lb_client.member_list(pool_id)
        except Exception as e:
            LOG.warning("Failed to communicate with load balancing service, "
                        "error: %s", str(e))
            raise threshold.InsufficientDataError(
                'failed to communicate with load balancing service',
                []
            )

        if getattr(ret, 'status_code', None):
            # Some error happened
            raise threshold.InsufficientDataError(ret.content, [])

        for m in ret.get("members", []):
            try:
                created_time = parser.parse(m['created_at'], ignoretz=True)
            except ValueError:
                LOG.warning('Failed to parse the member created time.')
                continue

            now = timeutils.utcnow()
            t = self.conf.member_creation_time
            if now - created_time < datetime.timedelta(seconds=t):
                LOG.debug("Ignore member which was created within %ss", t)
                continue

            if m["admin_state_up"] and m["operating_status"] == "ERROR":
                unhealthy_members.append(m)

        return unhealthy_members

    def _transition_alarm(self, alarm, new_state, members,
                          count, unknown_reason, pool_id=None,
                          stack_id=None, asg_id=None):
        transition = alarm.state != new_state
        last = members[-1] if members else None

        reason_data = {
            'type': ALARM_TYPE,
            'count': count,
            'most_recent': last,
            'unhealthy_members': members,
            "pool_id": pool_id,
            "stack_id": stack_id,
            "asg_id": asg_id
        }

        if transition:
            reason = ('Transition to %(state)s due to %(count)d members'
                      ' unhealthy, most recent: %(most_recent)s' %
                      dict(state=new_state, count=count, most_recent=last))
        else:
            reason = ('Remaining as %(state)s' % dict(state=new_state))

        reason = unknown_reason or reason

        # Refresh and trigger alarm based on state transition.
        self._refresh(alarm, new_state, reason, reason_data)

    def evaluate(self, alarm):
        if not self.within_time_constraint(alarm):
            LOG.debug('Attempted to evaluate alarm %s, but it is not '
                      'within its time constraint.', alarm.alarm_id)
            return

        LOG.debug("Evaluating %s rule alarm %s ...", ALARM_TYPE,
                  alarm.alarm_id)

        pool_id = alarm.rule["pool_id"]
        error_mems = []
        try:
            error_mems = self._get_unhealthy_members(pool_id)
        except threshold.InsufficientDataError as e:
            evaluation = (evaluator.UNKNOWN, e.statistics, 0, e.reason)
        else:
            state = evaluator.ALARM if len(error_mems) > 0 else evaluator.OK
            evaluation = (state, error_mems, len(error_mems), None)

        self._transition_alarm(alarm, *evaluation, pool_id=pool_id,
                               stack_id=alarm.rule.get("stack_id"),
                               asg_id=alarm.rule.get("autoscaling_group_id"))



#
# Copyright 2012 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import pecan

MEDIA_TYPE_JSON = 'application/vnd.openstack.telemetry-%s+json'
MEDIA_TYPE_XML = 'application/vnd.openstack.telemetry-%s+xml'


class VersionsController:

    @pecan.expose('json')
    def index(self):
        base_url = pecan.request.host_url
        available = [{'tag': 'v2', 'date': '2013-02-13T00:00:00Z', }]
        collected = [version_descriptor(base_url, v['tag'], v['date'])
                     for v in available]
        versions = {'versions': {'values': collected}}
        return versions


def version_descriptor(base_url, version, released_on):
    url = version_url(base_url, version)
    return {
        'id': version,
        'links': [
            {'href': url, 'rel': 'self', },
            {'href': 'https://docs.openstack.org/',
             'rel': 'describedby', 'type': 'text/html', }],
        'media-types': [
            {'base': 'application/json', 'type': MEDIA_TYPE_JSON % version, },
            {'base': 'application/xml', 'type': MEDIA_TYPE_XML % version, }],
        'status': 'stable',
        'updated': released_on,
    }


def version_url(base_url, version_number):
    return '{}/{}'.format(base_url, version_number)



# Copyright 2017 Fujitsu Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket

from oslo_log import log
from oslo_utils import importutils
import webob.dec

profiler = importutils.try_import('osprofiler.profiler')
profiler_initializer = importutils.try_import('osprofiler.initializer')
profiler_web = importutils.try_import('osprofiler.web')

LOG = log.getLogger(__name__)


class WsgiMiddleware:

    def __init__(self, application, **kwargs):
        self.application = application

    @classmethod
    def factory(cls, global_conf, **local_conf):
        if profiler_web:
            return profiler_web.WsgiMiddleware.factory(global_conf)

        def filter_(app):
            return cls(app)

        return filter_

    @webob.dec.wsgify
    def __call__(self, request):
        return request.get_response(self.application)


def setup(conf):
    if hasattr(conf, 'profiler') and conf.profiler.enabled:
        profiler_initializer.init_from_conf(
            conf=conf,
            context={},
            project=conf.project,
            service=conf.prog,
            host=socket.gethostbyname(socket.gethostname()))
        LOG.info('OSprofiler is enabled.')


def trace_cls(name, **kwargs):
    """Wrap the OSprofiler trace_cls.

    Wrap the OSprofiler trace_cls decorator so that it will not try to
    patch the class unless OSprofiler is present.

    :param name: The name of action. For example, wsgi, rpc, db, ...
    :param kwargs: Any other keyword args used by profiler.trace_cls
    """

    def decorator(cls):
        if profiler:
            trace_decorator = profiler.trace_cls(name, **kwargs)
            return trace_decorator(cls)
        return cls

    return decorator



# Copyright 2019 Catalyst Cloud Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from oslo_utils import netutils

from aodh.notifier import heat as heat_notifier
from aodh.tests.unit.notifier import base


class TestTrustHeatAlarmNotifier(base.TestNotifierBase):
    @mock.patch("aodh.keystone_client.get_heat_client_from_trust")
    def test_notify(self, mock_heatclient):
        action = netutils.urlsplit("trust+autohealer://fake_trust_id:delete@")
        alarm_id = "fake_alarm_id"
        alarm_name = "fake_alarm_name"
        severity = "low"
        previous = "ok"
        current = "alarm"
        reason = "no good reason"
        reason_data = {
            "stack_id": "fake_stack_id",
            "asg_id": "fake_asg_id",
            "unhealthy_members": [
                {"id": "3bd8bc5a-7632-11e9-84cd-00224d6b7bc1"}
            ]
        }

        class FakeResource:
            def __init__(self, resource_name):
                self.parent_resource = resource_name

        mock_client = mock_heatclient.return_value
        mock_client.resources.list.return_value = [
            FakeResource("fake_resource_name")
        ]

        notifier = heat_notifier.TrustHeatAlarmNotifier(self.conf)
        notifier.notify(action, alarm_id, alarm_name, severity, previous,
                        current, reason, reason_data)

        mock_heatclient.assert_called_once_with(self.conf, "fake_trust_id")
        mock_client.resources.mark_unhealthy.assert_called_once_with(
            "fake_asg_id",
            "fake_resource_name",
            True,
            "unhealthy load balancer member"
        )
        mock_client.stacks.update.assert_called_once_with(
            "fake_stack_id", existing=True
        )

    @mock.patch("aodh.keystone_client.get_heat_client_from_trust")
    def test_notify_stack_id_missing(self, mock_heatclient):
        action = netutils.urlsplit("trust+autohealer://fake_trust_id:delete@")
        alarm_id = "fake_alarm_id"
        alarm_name = "fake_alarm_name"
        severity = "low"
        previous = "ok"
        current = "alarm"
        reason = "no good reason"
        reason_data = {
            "asg_id": "fake_asg_id",
            "unhealthy_members": [
                {"tags": ["3bd8bc5a-7632-11e9-84cd-00224d6b7bc1"]}
            ]
        }

        notifier = heat_notifier.TrustHeatAlarmNotifier(self.conf)
        notifier.notify(action, alarm_id, alarm_name, severity, previous,
                        current, reason, reason_data)

        self.assertFalse(mock_heatclient.called)



#
# Copyright 2012 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_config import cfg

# Register options for the service
OPTS = [
    cfg.StrOpt('paste_config',
               default='api-paste.ini',
               help="Configuration file for WSGI definition of API."),
    cfg.StrOpt(
        'auth_mode',
        default="keystone",
        choices=('keystone', 'noauth'),
        help="Authentication mode to use."),
]



# Copyright 2020 Catalyst Cloud LTD.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from oslo_log import log
import pecan
from pecan import rest
import wsme
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from aodh.api.controllers.v2 import base
from aodh.api import rbac

LOG = log.getLogger(__name__)
ALLOWED_RESOURCES = ('alarms',)


class Quota(base.Base):
    resource = wtypes.wsattr(wtypes.Enum(str, *ALLOWED_RESOURCES),
                             mandatory=True)
    limit = wsme.wsattr(wtypes.IntegerType(minimum=-1), mandatory=True)


class Quotas(base.Base):
    project_id = wsme.wsattr(wtypes.text, mandatory=True)
    quotas = [Quota]


class QuotasController(rest.RestController):
    """Quota API controller."""

    @wsme_pecan.wsexpose(Quotas, str, ignore_extra_args=True)
    def get_all(self, project_id=None):
        """Get resource quotas of a project.

        - If no project given, get requested user's quota.
        - Admin user can get resource quotas of any project.
        """
        request_project = pecan.request.headers.get('X-Project-Id')
        project_id = project_id if project_id else request_project
        rbac.enforce(
            'get_quotas', pecan.request,
            pecan.request.enforcer, {'project_id': project_id})

        LOG.debug('Getting resource quotas for project %s', project_id)

        db_quotas = pecan.request.storage.get_quotas(project_id=project_id)

        if len(db_quotas) == 0:
            project_alarm_quota = pecan.request.cfg.api.project_alarm_quota
            quotas = [{'resource': 'alarms', 'limit': project_alarm_quota}]
            db_quotas = pecan.request.storage.set_quotas(project_id, quotas)

        quotas = [Quota.from_db_model(i) for i in db_quotas]
        return Quotas(project_id=project_id, quotas=quotas)

    @wsme_pecan.wsexpose(Quotas, body=Quotas, status_code=201)
    def post(self, body):
        """Create or update quota."""
        params = body.to_dict()
        project_id = params['project_id']

        rbac.enforce('update_quotas', pecan.request,
                     pecan.request.enforcer, {'project_id': project_id})

        input_quotas = []
        for i in params.get('quotas', []):
            input_quotas.append(i.to_dict())

        db_quotas = pecan.request.storage.set_quotas(project_id, input_quotas)
        quotas = [Quota.from_db_model(i) for i in db_quotas]

        return Quotas(project_id=project_id, quotas=quotas)

    @wsme_pecan.wsexpose(None, str, status_code=204)
    def delete(self, project_id):
        """Delete quotas for the given project."""
        rbac.enforce('delete_quotas', pecan.request,
                     pecan.request.enforcer, {'project_id': project_id})
        pecan.request.storage.delete_quotas(project_id)



#
# Copyright 2015 NEC Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time
from unittest import mock

from oslo_config import fixture as fixture_config
import oslo_messaging

from aodh import event
from aodh import service
from aodh.tests import base as tests_base


class TestEventAlarmEvaluationService(tests_base.BaseTestCase):

    def setUp(self):
        super().setUp()
        conf = service.prepare_service(argv=[], config_files=[])
        self.CONF = self.useFixture(fixture_config.Config(conf)).conf
        self.CONF.set_override("batch_size", 2, 'listener')
        self.setup_messaging(self.CONF)

    @mock.patch('aodh.storage.get_connection_from_config',
                mock.MagicMock())
    @mock.patch('aodh.event.EventAlarmEndpoint.sample')
    def test_batch_event_listener(self, mocked):
        msg_notifier = oslo_messaging.Notifier(
            self.transport, topics=['alarm.all'], driver='messaging',
            publisher_id='test-publisher')

        received_events = []
        mocked.side_effect = lambda msg: received_events.append(msg)
        event1 = {'event_type': 'compute.instance.update',
                  'traits': ['foo', 'bar'],
                  'message_id': '20d03d17-4aba-4900-a179-dba1281a3451',
                  'generated': '2016-04-23T06:50:21.622739'}
        event2 = {'event_type': 'compute.instance.update',
                  'traits': ['foo', 'bar'],
                  'message_id': '20d03d17-4aba-4900-a179-dba1281a3452',
                  'generated': '2016-04-23T06:50:23.622739'}
        msg_notifier.sample({}, 'event', event1)
        msg_notifier.sample({}, 'event', event2)

        svc = event.EventAlarmEvaluationService(0, self.CONF)
        self.addCleanup(svc.terminate)

        time.sleep(1)
        self.assertEqual(1, len(received_events))
        self.assertEqual(2, len(received_events[0]))



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Tests for aodh/evaluator/composite.py
"""
from unittest import mock

import fixtures
import os
from oslo_utils import timeutils
from oslo_utils import uuidutils

from aodh import evaluator
from aodh.evaluator import composite
from aodh.storage import models
from aodh.tests import constants
from aodh.tests.unit.evaluator import base


# NOTE(mmagr): Overriding PrometheusEvaluator setting to avoid
# complains during init.
os.environ['PROMETHEUS_HOST'] = '127.0.0.1'
os.environ['PROMETHEUS_PORT'] = '666'


class BaseCompositeEvaluate(base.TestEvaluatorBase):
    EVALUATOR = composite.CompositeEvaluator

    def setUp(self):
        self.client = self.useFixture(fixtures.MockPatch(
            'aodh.evaluator.gnocchi.client'
        )).mock.Client.return_value
        super().setUp()

    @staticmethod
    def _get_gnocchi_stats(granularity, values, aggregated=False):
        now = timeutils.utcnow_ts()
        if aggregated:
            return {
                'measures': {
                    'aggregated':
                        [[str(now - len(values) * granularity),
                          granularity, value] for value in values]
                }
            }
        return [[str(now - len(values) * granularity),
                 granularity, value] for value in values]

    @staticmethod
    def _reason(new_state, user_expression, causative_rules=(),
                transition=True):
        root_cause_rules = {}
        for index, rule in causative_rules:
            name = 'rule%s' % index
            root_cause_rules.update({name: rule})
        description = {evaluator.ALARM: 'outside their threshold.',
                       evaluator.OK: 'inside their threshold.',
                       evaluator.UNKNOWN: 'state evaluated to unknown.'}
        params = {'state': new_state,
                  'expression': user_expression,
                  'rules': ', '.join(sorted(root_cause_rules.keys())),
                  'description': description[new_state]}
        reason_data = {
            'type': 'composite',
            'composition_form': user_expression}
        reason_data.update(causative_rules=root_cause_rules)
        if transition:
            reason = ('Composite rule alarm with composition form: '
                      '%(expression)s transition to %(state)s, due to '
                      'rules: %(rules)s %(description)s' % params)
        else:
            reason = ('Composite rule alarm with composition form: '
                      '%(expression)s remaining as %(state)s, due to '
                      'rules: %(rules)s %(description)s' % params)
        return reason, reason_data


class CompositeTest(BaseCompositeEvaluate):
    sub_rule1 = {
        "type": "gnocchi_aggregation_by_metrics_threshold",
        "metrics": ['41869681-5776-46d6-91ed-cccc43b6e4e3',
                    'a1fb80f4-c242-4f57-87c6-68f47521059e'],
        "evaluation_periods": 5,
        "threshold": 0.8,
        "aggregation_method": "mean",
        "granularity": 60,
        "exclude_outliers": False,
        "comparison_operator": "gt"
    }

    sub_rule2 = {
        "type": "gnocchi_aggregation_by_metrics_threshold",
        "metrics": ['41869681-5776-46d6-91ed-cccc43b6e4e3',
                    'a1fb80f4-c242-4f57-87c6-68f47521059e'],
        "evaluation_periods": 4,
        "threshold": 200,
        "aggregation_method": "max",
        "granularity": 60,
        "exclude_outliers": False,
        "comparison_operator": "gt"
    }

    sub_rule3 = {
        "type": "gnocchi_aggregation_by_metrics_threshold",
        "metrics": ['41869681-5776-46d6-91ed-cccc43b6e4e3',
                    'a1fb80f4-c242-4f57-87c6-68f47521059e'],
        "evaluation_periods": 3,
        "threshold": 1000,
        "aggregation_method": "mean",
        "granularity": 60,
        "exclude_outliers": False,
        "comparison_operator": "gt"
    }

    sub_rule4 = {
        "type": "gnocchi_resources_threshold",
        'comparison_operator': 'gt',
        'threshold': 80.0,
        'evaluation_periods': 5,
        'aggregation_method': 'mean',
        'granularity': 60,
        'metric': 'cpu_util',
        'resource_type': 'instance',
        'resource_id': 'my_instance',
    }

    sub_rule5 = {
        "type": "gnocchi_aggregation_by_metrics_threshold",
        'comparison_operator': 'le',
        'threshold': 10.0,
        'evaluation_periods': 4,
        'aggregation_method': 'max',
        'granularity': 300,
        'metrics': ['0bb1604d-1193-4c0a-b4b8-74b170e35e83',
                    '9ddc209f-42f8-41e1-b8f1-8804f59c4053']
    }

    sub_rule6 = {
        "type": "gnocchi_aggregation_by_resources_threshold",
        'comparison_operator': 'gt',
        'threshold': 80.0,
        'evaluation_periods': 6,
        'aggregation_method': 'mean',
        'granularity': 50,
        'metric': 'cpu_util',
        'resource_type': 'instance',
        'query': '{"=": {"server_group": "my_autoscaling_group"}}'
    }

    def prepare_alarms(self):
        self.alarms = [
            models.Alarm(name='alarm_threshold_nest',
                         description='alarm with sub rules nested combined',
                         type='composite',
                         enabled=True,
                         user_id='fake_user',
                         project_id='fake_project',
                         alarm_id=uuidutils.generate_uuid(),
                         state='insufficient data',
                         state_reason='insufficient data',
                         state_timestamp=constants.MIN_DATETIME,
                         timestamp=constants.MIN_DATETIME,
                         insufficient_data_actions=[],
                         ok_actions=[],
                         alarm_actions=[],
                         repeat_actions=False,
                         time_constraints=[],
                         rule={
                             "or": [self.sub_rule1,
                                    {"and": [self.sub_rule2, self.sub_rule3]
                                     }]
                         },
                         severity='critical'),
            models.Alarm(name='alarm_threshold_or',
                         description='alarm on one of sub rules triggered',
                         type='composite',
                         enabled=True,
                         user_id='fake_user',
                         project_id='fake_project',
                         state='insufficient data',
                         state_reason='insufficient data',
                         state_timestamp=constants.MIN_DATETIME,
                         timestamp=constants.MIN_DATETIME,
                         insufficient_data_actions=[],
                         ok_actions=[],
                         alarm_actions=[],
                         repeat_actions=False,
                         alarm_id=uuidutils.generate_uuid(),
                         time_constraints=[],
                         rule={
                             "or": [self.sub_rule1, self.sub_rule2,
                                    self.sub_rule3]
                         },
                         severity='critical'
                         ),
            models.Alarm(name='alarm_threshold_and',
                         description='alarm on all the sub rules triggered',
                         type='composite',
                         enabled=True,
                         user_id='fake_user',
                         project_id='fake_project',
                         state='insufficient data',
                         state_reason='insufficient data',
                         state_timestamp=constants.MIN_DATETIME,
                         timestamp=constants.MIN_DATETIME,
                         insufficient_data_actions=[],
                         ok_actions=[],
                         alarm_actions=[],
                         repeat_actions=False,
                         alarm_id=uuidutils.generate_uuid(),
                         time_constraints=[],
                         rule={
                             "and": [self.sub_rule1, self.sub_rule2,
                                     self.sub_rule3]
                         },
                         severity='critical'
                         ),
            models.Alarm(name='alarm_multi_type_rules',
                         description='alarm with threshold and gnocchi rules',
                         type='composite',
                         enabled=True,
                         user_id='fake_user',
                         project_id='fake_project',
                         alarm_id=uuidutils.generate_uuid(),
                         state='insufficient data',
                         state_reason='insufficient data',
                         state_timestamp=constants.MIN_DATETIME,
                         timestamp=constants.MIN_DATETIME,
                         insufficient_data_actions=[],
                         ok_actions=[],
                         alarm_actions=[],
                         repeat_actions=False,
                         time_constraints=[],
                         rule={
                             "and": [self.sub_rule2, self.sub_rule3,
                                     {'or': [self.sub_rule1, self.sub_rule4,
                                             self.sub_rule5, self.sub_rule6]}]
                         },
                         severity='critical'
                         ),
        ]

    def test_simple_insufficient(self):
        self._set_all_alarms('ok')
        self.client.aggregates.fetch.return_value = []
        self.client.metric.get_measures.return_value = []
        self._evaluate_all_alarms()
        self._assert_all_alarms('insufficient data')
        expected = [mock.call(alarm) for alarm in self.alarms]
        update_calls = self.storage_conn.update_alarm.call_args_list
        self.assertEqual(expected, update_calls)
        expected = [mock.call(self.alarms[0],
                              'ok',
                              *self._reason(
                                  'insufficient data',
                                  '(rule1 or (rule2 and rule3))',
                                  ((1, self.sub_rule1), (2, self.sub_rule2),
                                   (3, self.sub_rule3)))),
                    mock.call(self.alarms[1],
                              'ok',
                              *self._reason(
                                  'insufficient data',
                                  '(rule1 or rule2 or rule3)',
                                  ((1, self.sub_rule1), (2, self.sub_rule2),
                                   (3, self.sub_rule3)))),
                    mock.call(self.alarms[2],
                              'ok',
                              *self._reason(
                                  'insufficient data',
                                  '(rule1 and rule2 and rule3)',
                                  ((1, self.sub_rule1), (2, self.sub_rule2),
                                   (3, self.sub_rule3)))),
                    mock.call(
                        self.alarms[3],
                        'ok',
                        *self._reason(
                            'insufficient data',
                            '(rule1 and rule2 and (rule3 or rule4 or rule5 '
                            'or rule6))',
                            ((1, self.sub_rule2), (2, self.sub_rule3),
                             (3, self.sub_rule1), (4, self.sub_rule4),
                             (5, self.sub_rule5), (6, self.sub_rule6))))]
        self.assertEqual(expected, self.notifier.notify.call_args_list)

    def test_alarm_full_trip_with_multi_type_rules(self):
        alarm = self.alarms[3]
        alarm.state = 'ok'
        # following results of sub-rules evaluation to trigger
        # final "alarm" state:
        # self.sub_rule2: alarm
        # self.sub_rule3: alarm
        # self.sub_rule1: ok
        # self.sub_rule4: ok
        # self.sub_rule5: ok
        # self.sub_rule6: alarm
        maxs = self._get_gnocchi_stats(
            60, [self.sub_rule2['threshold'] + v
                 for v in range(1, 5)],
            aggregated=True)
        avgs1 = self._get_gnocchi_stats(
            60, [self.sub_rule3['threshold'] + v
                 for v in range(1, 4)])
        avgs2 = self._get_gnocchi_stats(
            60, [self.sub_rule1['threshold'] - v
                 for v in range(1, 6)],
            aggregated=True)
        gavgs1 = self._get_gnocchi_stats(
            60, [self.sub_rule4['threshold']
                 - v for v in range(1, 6)],
            aggregated=True)
        gmaxs = self._get_gnocchi_stats(
            300, [self.sub_rule5['threshold'] + v
                  for v in range(1, 5)],
            aggregated=True)
        gavgs2 = self._get_gnocchi_stats(
            50, [self.sub_rule6['threshold'] + v
                 for v in range(1, 7)],
            aggregated=True)

        self.client.metric.get_measures.side_effect = [gavgs1]
        self.client.aggregates.fetch.side_effect = [maxs, avgs1, avgs2,
                                                    gmaxs, gavgs2]
        self.evaluator.evaluate(alarm)
        self.assertEqual(1, self.client.metric.get_measures.call_count)
        self.assertEqual(5, self.client.aggregates.fetch.call_count)
        self.assertEqual('alarm', alarm.state)
        expected = mock.call(
            alarm, 'ok',
            *self._reason(
                'alarm',
                '(rule1 and rule2 and (rule3 or rule4 or rule5 or rule6))',
                ((1, self.sub_rule2), (2, self.sub_rule3),
                 (6, self.sub_rule6))))
        self.assertEqual(expected, self.notifier.notify.call_args)

    def test_alarm_with_short_circuit_logic(self):
        alarm = self.alarms[1]
        # self.sub_rule1: alarm
        avgs = self._get_gnocchi_stats(
            60, [self.sub_rule1['threshold'] + v
                 for v in range(1, 6)],
            aggregated=True)
        self.client.aggregates.fetch.side_effect = [avgs]
        self.evaluator.evaluate(alarm)
        self.assertEqual('alarm', alarm.state)
        self.assertEqual(1, self.client.aggregates.fetch.call_count)
        expected = mock.call(self.alarms[1], 'insufficient data',
                             *self._reason(
                                 'alarm',
                                 '(rule1 or rule2 or rule3)',
                                 ((1, self.sub_rule1),)))
        self.assertEqual(expected, self.notifier.notify.call_args)

    def test_ok_with_short_circuit_logic(self):
        alarm = self.alarms[2]
        # self.sub_rule1: ok
        avgs = self._get_gnocchi_stats(
            60, [self.sub_rule1['threshold'] - v
                 for v in range(1, 6)],
            aggregated=True)
        self.client.aggregates.fetch.side_effect = [avgs]
        self.evaluator.evaluate(alarm)
        self.assertEqual('ok', alarm.state)
        self.assertEqual(1, self.client.aggregates.fetch.call_count)
        expected = mock.call(self.alarms[2], 'insufficient data',
                             *self._reason(
                                 'ok',
                                 '(rule1 and rule2 and rule3)',
                                 ((1, self.sub_rule1),)))
        self.assertEqual(expected, self.notifier.notify.call_args)

    def test_unknown_state_with_sub_rules_trending_state(self):
        alarm = self.alarms[0]
        maxs = self._get_gnocchi_stats(
            60, [self.sub_rule2['threshold'] + v
                 for v in range(-1, 4)],
            aggregated=True)
        avgs = self._get_gnocchi_stats(
            60, [self.sub_rule3['threshold'] + v
                 for v in range(-1, 3)],
            aggregated=True)
        avgs2 = self._get_gnocchi_stats(
            60, [self.sub_rule1['threshold'] - v
                 for v in range(1, 6)],
            aggregated=True)
        self.client.aggregates.fetch.side_effect = [avgs2, maxs, avgs]

        self.evaluator.evaluate(alarm)
        self.assertEqual('alarm', alarm.state)
        expected = mock.call(self.alarms[0], 'insufficient data',
                             *self._reason(
                                 'alarm',
                                 '(rule1 or (rule2 and rule3))',
                                 ((2, self.sub_rule2),
                                  (3, self.sub_rule3))))
        self.assertEqual(expected, self.notifier.notify.call_args)

    def test_known_state_with_sub_rules_trending_state(self):
        alarm = self.alarms[0]
        alarm.repeat_actions = True
        alarm.state = 'ok'

        maxs = self._get_gnocchi_stats(
            60, [self.sub_rule2['threshold'] + v
                 for v in range(-1, 4)],
            aggregated=True)
        avgs = self._get_gnocchi_stats(
            60, [self.sub_rule3['threshold'] + v
                 for v in range(-1, 3)],
            aggregated=True)
        avgs2 = self._get_gnocchi_stats(
            60, [self.sub_rule1['threshold'] - v
                 for v in range(1, 6)],
            aggregated=True)
        self.client.aggregates.fetch.side_effect = [avgs2, maxs, avgs]

        self.evaluator.evaluate(alarm)
        self.assertEqual('ok', alarm.state)
        expected = mock.call(self.alarms[0], 'ok',
                             *self._reason(
                                 'ok',
                                 '(rule1 or (rule2 and rule3))',
                                 ((1, self.sub_rule1),
                                  (2, self.sub_rule2),
                                  (3, self.sub_rule3)), False))
        self.assertEqual(expected, self.notifier.notify.call_args)

    def test_known_state_with_sub_rules_trending_state_and_not_repeat(self):
        alarm = self.alarms[2]
        alarm.state = 'ok'
        maxs = self._get_gnocchi_stats(
            60, [self.sub_rule2['threshold'] + v
                 for v in range(-1, 4)],
            aggregated=True)
        avgs = self._get_gnocchi_stats(
            60, [self.sub_rule3['threshold'] + v
                 for v in range(-1, 3)],
            aggregated=True)
        avgs2 = self._get_gnocchi_stats(
            60, [self.sub_rule1['threshold'] - v
                 for v in range(1, 6)],
            aggregated=True)
        self.client.aggregates.fetch.side_effect = [avgs2, maxs, avgs]
        self.evaluator.evaluate(alarm)
        self.assertEqual('ok', alarm.state)
        self.assertEqual([], self.notifier.notify.mock_calls)


class OtherCompositeTest(BaseCompositeEvaluate):
    sub_rule1 = {
        'evaluation_periods': 3,
        'metric': 'radosgw.objects.containers',
        'resource_id': 'alarm-resource-1',
        'aggregation_method': 'mean',
        'granularity': 60,
        'threshold': 5.0,
        'type': 'gnocchi_resources_threshold',
        'comparison_operator': 'ge',
        'resource_type': 'ceph_account'
    }

    sub_rule2 = {
        'evaluation_periods': 3,
        'metric': 'radosgw.objects.containers',
        'resource_id': 'alarm-resource-2',
        'aggregation_method': 'mean',
        'granularity': 60,
        'threshold': 5.0,
        'type': 'gnocchi_resources_threshold',
        'comparison_operator': 'ge',
        'resource_type': 'ceph_account'
    }

    def prepare_alarms(self):
        self.alarms = [
            models.Alarm(name='composite-GRT-OR-GRT',
                         description='composite alarm converted',
                         type='composite',
                         enabled=True,
                         user_id='fake_user',
                         project_id='fake_project',
                         state='insufficient data',
                         state_reason='insufficient data',
                         state_timestamp=constants.MIN_DATETIME,
                         timestamp=constants.MIN_DATETIME,
                         insufficient_data_actions=['log://'],
                         ok_actions=['log://'],
                         alarm_actions=['log://'],
                         repeat_actions=False,
                         alarm_id=uuidutils.generate_uuid(),
                         time_constraints=[],
                         rule={
                             "or": [self.sub_rule1, self.sub_rule2]
                         },
                         severity='critical'
                         ),
        ]

    def test_simple_ok(self):
        self._set_all_alarms('alarm')

        gavgs1 = [['2016-11-24T10:00:00+00:00', 3600.0, 3.0],
                  ['2016-11-24T10:00:00+00:00', 900.0, 3.0],
                  ['2016-11-24T10:00:00+00:00', 300.0, 3.0],
                  ['2016-11-24T10:01:00+00:00', 60.0, 2.0],
                  ['2016-11-24T10:02:00+00:00', 60.0, 3.0],
                  ['2016-11-24T10:03:00+00:00', 60.0, 4.0],
                  ['2016-11-24T10:04:00+00:00', 60.0, 5.0]]

        gavgs2 = [['2016-11-24T10:00:00+00:00', 3600.0, 3.0],
                  ['2016-11-24T10:00:00+00:00', 900.0, 3.0],
                  ['2016-11-24T10:00:00+00:00', 300.0, 3.0],
                  ['2016-11-24T10:01:00+00:00', 60.0, 2.0],
                  ['2016-11-24T10:02:00+00:00', 60.0, 3.0],
                  ['2016-11-24T10:03:00+00:00', 60.0, 4.0],
                  ['2016-11-24T10:04:00+00:00', 60.0, 5.0]]

        self.client.metric.get_measures.side_effect = [gavgs1, gavgs2]
        self._evaluate_all_alarms()
        self._assert_all_alarms('ok')
        expected = [mock.call(alarm) for alarm in self.alarms]
        update_calls = self.storage_conn.update_alarm.call_args_list
        self.assertEqual(expected, update_calls)
        expected = [mock.call(self.alarms[0], 'alarm',
                              *self._reason('ok', '(rule1 or rule2)',
                                            ((1, self.sub_rule1),
                                             (2, self.sub_rule2))))]
        self.assertEqual(expected, self.notifier.notify.call_args_list)



#
# Copyright 2013 IBM Corp.
# Copyright 2013 Julien Danjou
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Test basic aodh-api app
"""
import json
from unittest import mock

import wsme

from aodh import i18n
from aodh.tests.functional.api import v2


class TestApiMiddleware(v2.FunctionalTest):

    no_lang_translated_error = 'No lang translated error'
    en_US_translated_error = 'en-US translated error'

    def _fake_translate(self, message, user_locale):
        if user_locale is None:
            return self.no_lang_translated_error
        else:
            return self.en_US_translated_error

    def test_json_parsable_error_middleware_translation_400(self):
        # Ensure translated messages get placed properly into json faults
        with mock.patch.object(i18n, 'translate',
                               side_effect=self._fake_translate):
            response = self.post_json('/alarms', params={
                'name': 'foobar', 'type': 'gnocchi_resources_threshold'},
                expect_errors=True, headers={"Accept": "application/json"})
        self.assertEqual(400, response.status_int)
        self.assertEqual("application/json", response.content_type)
        self.assertTrue(response.json['error_message'])
        self.assertEqual(self.no_lang_translated_error,
                         response.json['error_message']['faultstring'])

    def test_xml_parsable_error_middleware_translation_400(self):
        # Ensure translated messages get placed properly into xml faults
        with mock.patch.object(i18n, 'translate',
                               side_effect=self._fake_translate):
            response = self.post_json('/alarms', params={
                'name': 'foobar', 'type': 'gnocchi_resources_threshold'},
                expect_errors=True, headers={"Accept": "application/xml,*/*"})
        self.assertEqual(400, response.status_int)
        self.assertEqual("application/xml", response.content_type)
        self.assertEqual('error_message', response.xml.tag)
        fault = response.xml.findall('./error/faultstring')
        for fault_string in fault:
            self.assertEqual(self.no_lang_translated_error, fault_string.text)

    def test_best_match_language(self):
        # Ensure that we are actually invoking language negotiation
        with mock.patch.object(i18n, 'translate',
                               side_effect=self._fake_translate):
            response = self.post_json('/alarms', params={
                'name': 'foobar', 'type': 'gnocchi_resources_threshold'},
                expect_errors=True, headers={"Accept": "application/xml,*/*",
                                             "Accept-Language": "en-US"})

        self.assertEqual(400, response.status_int)
        self.assertEqual("application/xml", response.content_type)
        self.assertEqual('error_message', response.xml.tag)
        fault = response.xml.findall('./error/faultstring')
        for fault_string in fault:
            self.assertEqual(self.en_US_translated_error, fault_string.text)

    def test_translated_then_untranslated_error(self):
        resp = self.get_json('/alarms/alarm-id-3', expect_errors=True)
        self.assertEqual(404, resp.status_code)
        body = resp.body.decode('utf-8')
        self.assertEqual("Alarm alarm-id-3 not found",
                         json.loads(body)['error_message']
                         ['faultstring'])

        with mock.patch('aodh.api.controllers.'
                        'v2.base.AlarmNotFound') as CustomErrorClass:
            CustomErrorClass.return_value = wsme.exc.ClientSideError(
                "untranslated_error", status_code=404)
            resp = self.get_json('/alarms/alarm-id-5', expect_errors=True)

        self.assertEqual(404, resp.status_code)
        body = resp.body.decode('utf-8')
        self.assertEqual("untranslated_error",
                         json.loads(body)['error_message']
                         ['faultstring'])



# Copyright 2025 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""added_counter_table

Revision ID: 008
Revises: 007
Create Date: 2025-01-15 10:28:02.087788

"""

# revision identifiers, used by Alembic.
revision = '008'
down_revision = '007'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'alarm_counter',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('alarm_id', sa.String(length=128), nullable=False),
        sa.Column('project_id', sa.String(length=128), nullable=False),
        sa.Column('state', sa.String(length=128), nullable=False),
        sa.Column('value', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ['alarm_id'],
            ['alarm.alarm_id'],
            name='alarm_fkey_ref',
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('alarm_id', 'project_id', 'state')
    )
    op.create_index(
        'ix_alarm_counter_alarm_id',
        'alarm_counter',
        ['alarm_id'],
        unique=False
    )
    op.create_index(
        'ix_alarm_counter_project_id',
        'alarm_counter',
        ['project_id'],
        unique=False
    )
    op.create_index(
        'ix_alarm_counter_state',
        'alarm_counter',
        ['state'],
        unique=False
    )



# Copyright 2019 Catalyst Cloud Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime
from unittest import mock

from oslo_utils import timeutils
from oslo_utils import uuidutils

from aodh import evaluator
from aodh.evaluator import loadbalancer
from aodh.storage import models
from aodh.tests import constants
from aodh.tests.unit.evaluator import base


@mock.patch('octaviaclient.api.v2.octavia.OctaviaAPI')
@mock.patch('aodh.keystone_client.get_session')
class TestLoadBalancerMemberHealthEvaluator(base.TestEvaluatorBase):
    EVALUATOR = loadbalancer.LoadBalancerMemberHealthEvaluator

    def test_evaluate(self, mock_session, mock_octavia):
        alarm = models.Alarm(
            name='lb_member_alarm',
            description='lb_member_alarm',
            type=loadbalancer.ALARM_TYPE,
            enabled=True,
            user_id=uuidutils.generate_uuid(),
            project_id=uuidutils.generate_uuid(dashed=False),
            alarm_id=uuidutils.generate_uuid(),
            state='insufficient data',
            state_reason='insufficient data',
            state_timestamp=constants.MIN_DATETIME,
            timestamp=constants.MIN_DATETIME,
            insufficient_data_actions=[],
            ok_actions=[],
            alarm_actions=[],
            repeat_actions=False,
            time_constraints=[],
            severity='low',
            rule=dict(
                pool_id=uuidutils.generate_uuid(),
                stack_id=uuidutils.generate_uuid(),
                autoscaling_group_id=uuidutils.generate_uuid(),
            )
        )

        mock_client = mock.MagicMock()
        mock_octavia.return_value = mock_client
        created_at = timeutils.utcnow() - datetime.timedelta(days=1)
        mock_client.member_list.return_value = {
            'members': [
                {
                    'created_at': created_at.isoformat(),
                    'admin_state_up': True,
                    'operating_status': 'ERROR',
                }
            ]
        }

        self.evaluator.evaluate(alarm)

        self.assertEqual(evaluator.ALARM, alarm.state)

    def test_evaluate_octavia_error(self, mock_session, mock_octavia):
        class Response:
            def __init__(self, status_code, content):
                self.status_code = status_code
                self.content = content

        alarm = models.Alarm(
            name='lb_member_alarm',
            description='lb_member_alarm',
            type=loadbalancer.ALARM_TYPE,
            enabled=True,
            user_id=uuidutils.generate_uuid(),
            project_id=uuidutils.generate_uuid(dashed=False),
            alarm_id=uuidutils.generate_uuid(),
            state='insufficient data',
            state_reason='insufficient data',
            state_timestamp=constants.MIN_DATETIME,
            timestamp=constants.MIN_DATETIME,
            insufficient_data_actions=[],
            ok_actions=[],
            alarm_actions=[],
            repeat_actions=False,
            time_constraints=[],
            severity='low',
            rule=dict(
                pool_id=uuidutils.generate_uuid(),
                stack_id=uuidutils.generate_uuid(),
                autoscaling_group_id=uuidutils.generate_uuid(),
            )
        )

        mock_client = mock.MagicMock()
        mock_octavia.return_value = mock_client
        msg = 'Pool NotFound'
        mock_client.member_list.return_value = Response(404, msg)

        self.evaluator.evaluate(alarm)

        self.assertEqual(evaluator.UNKNOWN, alarm.state)
        self.assertEqual(msg, alarm.state_reason)

    def test_evaluate_alarm_to_ok(self, mock_session, mock_octavia):
        alarm = models.Alarm(
            name='lb_member_alarm',
            description='lb_member_alarm',
            type=loadbalancer.ALARM_TYPE,
            enabled=True,
            user_id=uuidutils.generate_uuid(),
            project_id=uuidutils.generate_uuid(dashed=False),
            alarm_id=uuidutils.generate_uuid(),
            state=evaluator.ALARM,
            state_reason='alarm',
            state_timestamp=constants.MIN_DATETIME,
            timestamp=constants.MIN_DATETIME,
            insufficient_data_actions=[],
            ok_actions=[],
            alarm_actions=[],
            repeat_actions=False,
            time_constraints=[],
            severity='low',
            rule=dict(
                pool_id=uuidutils.generate_uuid(),
                stack_id=uuidutils.generate_uuid(),
                autoscaling_group_id=uuidutils.generate_uuid(),
            )
        )

        mock_client = mock.MagicMock()
        mock_octavia.return_value = mock_client
        created_at = timeutils.utcnow() - datetime.timedelta(days=1)
        mock_client.member_list.return_value = {
            'members': [
                {
                    'created_at': created_at.isoformat(),
                    'admin_state_up': True,
                    'operating_status': 'ACTIVE',
                }
            ]
        }

        self.evaluator.evaluate(alarm)

        self.assertEqual(evaluator.OK, alarm.state)



#
# Copyright 2013 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Model classes for use in the storage API.
"""

import datetime

from aodh.i18n import _
from aodh.storage import base


class Alarm(base.Model):
    ALARM_INSUFFICIENT_DATA = 'insufficient data'
    ALARM_OK = 'ok'
    ALARM_ALARM = 'alarm'

    ALARM_ACTIONS_MAP = {
        ALARM_INSUFFICIENT_DATA: 'insufficient_data_actions',
        ALARM_OK: 'ok_actions',
        ALARM_ALARM: 'alarm_actions',
    }

    ALARM_LEVEL_LOW = 'low'
    ALARM_LEVEL_MODERATE = 'moderate'
    ALARM_LEVEL_CRITICAL = 'critical'

    SUPPORT_SORT_KEYS = (
        'alarm_id', 'enabled', 'name', 'type', 'severity', 'timestamp',
        'user_id', 'project_id', 'state', 'repeat_actions', 'state_timestamp')
    DEFAULT_SORT = [('timestamp', 'desc')]
    PRIMARY_KEY = 'alarm_id'

    """
    An alarm to monitor.

    :param alarm_id: UUID of the alarm
    :param type: type of the alarm
    :param name: The Alarm name
    :param description: User friendly description of the alarm
    :param enabled: Is the alarm enabled
    :param state: Alarm state (ok/alarm/insufficient data)
    :param state_reason: Alarm state reason
    :param rule: A rule that defines when the alarm fires
    :param user_id: the owner/creator of the alarm
    :param project_id: the project_id of the creator
    :param evaluation_periods: the number of periods
    :param period: the time period in seconds
    :param time_constraints: the list of the alarm's time constraints, if any
    :param timestamp: the timestamp when the alarm was last updated
    :param state_timestamp: the timestamp of the last state change
    :param ok_actions: the list of webhooks to call when entering the ok state
    :param alarm_actions: the list of webhooks to call when entering the
                          alarm state
    :param insufficient_data_actions: the list of webhooks to call when
                                      entering the insufficient data state
    :param repeat_actions: Is the actions should be triggered on each
                           alarm evaluation.
    :param severity: Alarm level (low/moderate/critical)
    :param evaluate_timestamp: The timestamp when the alarm is finished
                               evaluating.
    """
    def __init__(self, alarm_id, type, enabled, name, description,
                 timestamp, user_id, project_id, state, state_timestamp,
                 state_reason, ok_actions, alarm_actions,
                 insufficient_data_actions, repeat_actions, rule,
                 time_constraints, severity=None, evaluate_timestamp=None):
        if not isinstance(timestamp, datetime.datetime):
            raise TypeError(_("timestamp should be datetime object"))
        if not isinstance(state_timestamp, datetime.datetime):
            raise TypeError(_("state_timestamp should be datetime object"))
        super().__init__(
            alarm_id=alarm_id,
            type=type,
            enabled=enabled,
            name=name,
            description=description,
            timestamp=timestamp,
            user_id=user_id,
            project_id=project_id,
            state=state,
            state_timestamp=state_timestamp,
            state_reason=state_reason,
            ok_actions=ok_actions,
            alarm_actions=alarm_actions,
            insufficient_data_actions=insufficient_data_actions,
            repeat_actions=repeat_actions,
            rule=rule,
            time_constraints=time_constraints,
            severity=severity,
            evaluate_timestamp=evaluate_timestamp)


class AlarmChange(base.Model):
    """Record of an alarm change.

    :param event_id: UUID of the change event
    :param alarm_id: UUID of the alarm
    :param type: The type of change
    :param severity: The severity of alarm
    :param detail: JSON fragment describing change
    :param user_id: the user ID of the initiating identity
    :param project_id: the project ID of the initiating identity
    :param on_behalf_of: the tenant on behalf of which the change
                         is being made
    :param timestamp: the timestamp of the change
    """

    CREATION = 'creation'
    RULE_CHANGE = 'rule change'
    STATE_TRANSITION = 'state transition'
    DELETION = 'deletion'

    SUPPORT_SORT_KEYS = (
        'event_id', 'alarm_id', 'on_behalf_of', 'project_id', 'user_id',
        'type', 'timestamp', 'severity')
    DEFAULT_SORT = [('timestamp', 'desc')]
    PRIMARY_KEY = 'event_id'

    def __init__(self,
                 event_id,
                 alarm_id,
                 type,
                 detail,
                 user_id,
                 project_id,
                 on_behalf_of,
                 severity=None,
                 timestamp=None
                 ):
        super().__init__(
            event_id=event_id,
            alarm_id=alarm_id,
            type=type,
            severity=severity,
            detail=detail,
            user_id=user_id,
            project_id=project_id,
            on_behalf_of=on_behalf_of,
            timestamp=timestamp)


class Quota(base.Model):
    def __init__(self, project_id, resource, limit):
        super().__init__(
            project_id=project_id,
            resource=resource,
            limit=limit)


class AlarmCounter(base.Model):
    def __init__(self, alarm_id, project_id, state):
        super().__init__(
            alarm_id=alarm_id,
            project_id=project_id,
            state=state,
            value=0
        )



# Copyright 2025 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Fix missing FK constraint

Revision ID: 009
Revises: 008
Create Date: 2025-07-18 23:45:50.411424

"""

# revision identifiers, used by Alembic.
revision = '009'
down_revision = '008'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('alarm_history', 'alarm_id', nullable=False,
                    existing_type=sa.String(128), existing_nullable=True,
                    existing_server_default=False)
    op.create_foreign_key(
        constraint_name=None,
        source_table='alarm_history', referent_table='alarm',
        local_cols=['alarm_id'], remote_cols=['alarm_id'])



# Copyright 2019 Catalyst Cloud Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log

from aodh import keystone_client as aodh_keystone
from aodh import notifier

LOG = log.getLogger(__name__)


class TrustHeatAlarmNotifier(notifier.AlarmNotifier):
    """Heat autohealing notifier.

    The auto-healing notifier works together with loadbalancer_member_health
    evaluator.

    Presumably, the end user defines a Heat template which contains an
    autoscaling group and all the members in the group are joined in an Octavia
    load balancer in order to expose service to the outside, so that when the
    stack scales up or scales down, Heat makes sure the new members are joining
    the load balancer automatically and the old members are removed.

    However, this notifier deals with the situation that when some member
    fails, the stack could be recovered by marking the given autoscaling group
    member unhealthy, then update Heat stack in place. In order to do that, the
    notifier needs to know:

    - Heat top/root stack ID.
    - Heat autoscaling group ID.
    - The failed Octavia pool members.
    """

    def __init__(self, conf):
        super().__init__(conf)
        self.conf = conf

    def notify(self, action, alarm_id, alarm_name, severity, previous, current,
               reason, reason_data):
        LOG.info(
            "Notifying alarm %(alarm_name)s %(alarm_id)s of %(severity)s "
            "priority from %(previous)s to %(current)s with action %(action)s"
            " because %(reason)s.",
            {'alarm_name': alarm_name,
             'alarm_id': alarm_id,
             'severity': severity,
             'previous': previous,
             'current': current,
             'action': action.geturl(),
             'reason': reason}
        )

        trust_id = action.username
        stack_id = reason_data.get("stack_id")
        asg_id = reason_data.get("asg_id")
        unhealthy_members = reason_data.get("unhealthy_members", [])
        unhealthy_resources = []

        if not stack_id or not asg_id:
            LOG.error(
                "stack_id and asg_id must exist to notify alarm %s", alarm_id
            )
            return

        heat_client = aodh_keystone.get_heat_client_from_trust(
            self.conf, trust_id
        )

        for member in unhealthy_members:
            target_resources = heat_client.resources.list(
                stack_id, nested_depth=3,
                filters={"physical_resource_id": member["id"]}
            )
            if len(target_resources) > 0:
                # There should be only one item.
                unhealthy_resources.append(
                    target_resources[0].parent_resource
                )

        if not unhealthy_resources:
            LOG.warning("No unhealthy resource found for the alarm %s",
                        alarm_id)
            return

        try:
            for res in unhealthy_resources:
                heat_client.resources.mark_unhealthy(
                    asg_id,
                    res,
                    True,
                    "unhealthy load balancer member"
                )
                LOG.info(
                    "Heat resource %(resource_id)s is marked as unhealthy "
                    "for alarm %(alarm_id)s",
                    {"resource_id": res, "alarm_id": alarm_id}
                )

            heat_client.stacks.update(stack_id, existing=True)
            LOG.info(
                "Heat stack %(stack_id)s is updated for alarm "
                "%(alarm_id)s",
                {"stack_id": stack_id, "alarm_id": alarm_id}
            )
        except Exception as e:
            LOG.exception("Failed to communicate with Heat service for alarm "
                          "%s, error: %s",
                          alarm_id, str(e))



#
# Copyright 2023 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log
import wsme
from wsme import types as wtypes

from aodh.api.controllers.v2 import base


LOG = log.getLogger(__name__)


class PrometheusRule(base.AlarmRule):
    comparison_operator = base.AdvEnum('comparison_operator', str,
                                       'lt', 'le', 'eq', 'ne', 'ge', 'gt',
                                       default='eq')
    "The comparison against the alarm threshold"

    threshold = wsme.wsattr(float, mandatory=True)
    "The threshold of the alarm"

    query = wsme.wsattr(wtypes.text, mandatory=True)
    "The Prometheus query"

    @staticmethod
    def validate(rule):
        # TO-DO(mmagr): validate Prometheus query maybe?
        return rule

    def as_dict(self):
        rule = self.as_dict_from_keys(['comparison_operator', 'threshold',
                                       'query'])
        return rule
